@@ -44,7 +44,7 @@ static char iblkbuf[EXT2_MAX_BLOCK_SIZE];
 static int cached_diblkno = -1;
 static char diblkbuf[EXT2_MAX_BLOCK_SIZE];
 static long dev = -1;
-static long partition_offset;
+static __u64 partition_offset;
 
 struct ext2_inode *fd2inode[MAX_FD];
 
@@ -64,6 +64,27 @@ static struct inode_table_entry {
 #undef DEBUG
 #define Debug 0
 
+/* div_blocksize() and mul_blocksize() are trivial functions to avoid 64-bit
+   __udivdi3 and __muldi3 millicode routines. They only work because the ext2
+   blocksize is a multiple of 2 */
+static __u64 div_blocksize(__u64 devaddr, unsigned blocksize)
+{
+	while (blocksize > 1) {
+		devaddr >>= 1;
+		blocksize >>= 1;
+	}
+	return devaddr;
+}
+static __u64 mul_blocksize(__u64 devaddr, unsigned blocksize)
+{
+	while (blocksize > 1) {
+		devaddr <<= 1;
+		blocksize >>= 1;
+	}
+	return devaddr;
+}
+
+
 static struct ext3_extent_header *ext3_extent_header(struct ext2_inode *i)
 {
 	return (struct ext3_extent_header *)&i->i_block[0];
@@ -209,8 +230,12 @@ static void swapino(struct ext2_inode *i)
 #endif
 }
 
-static void swapde(struct ext2_dir_entry_2 *de)
+static struct ext2_dir_entry_2 de_temp;
+static struct ext2_dir_entry_2 *swapde(struct ext2_dir_entry_2 *source)
 {
+    struct ext2_dir_entry_2 *de = &de_temp;
+
+    memcpy(de, source, sizeof(de_temp));
     if (Debug) printf("swapde: 0x%x 0x%x / ",
 				de->inode, de->rec_len);
     inplace(__le32_to_cpu, de->inode);
@@ -219,6 +244,7 @@ static void swapde(struct ext2_dir_entry_2 *de)
 				de->inode, de->rec_len,
 				de->name_len, de->file_type,
 				de->name_len, de->name);
+    return de;
 }
 
 static struct ext2_group_desc *ext2_gds(int i)
@@ -262,7 +288,7 @@ int ext2_mount(long cons_dev, long p_offset, long quiet)
 	
 	if (__le16_to_cpu(sb.s_magic) != EXT2_SUPER_MAGIC) {
 		if (!quiet) {
-			printf("ext2_init: bad magic 0x%x\n", sb.s_magic);
+			printf("This is not a ext2 filesystem (magic 0x%x)\n", sb.s_magic);
 		}
 		return -1;
 	}
@@ -333,7 +359,7 @@ static struct ext2_inode *ext2_iget(int ino)
 	struct ext2_inode *ip;
 	struct inode_table_entry *itp = 0;
 	int group;
-	long offset;
+	__u64 offset;
 
 
 	ip = 0;
@@ -358,12 +384,12 @@ static struct ext2_inode *ext2_iget(int ino)
 	printf("group is %d\n", group);
 #endif
 	offset = partition_offset
-		+ ((long) ext2_gds(group)->bg_inode_table * (long)ext2_blocksize)
+		+ mul_blocksize(ext2_gds(group)->bg_inode_table, ext2_blocksize)
 		+ (((ino - 1) % EXT2_INODES_PER_GROUP(&sb))
 		   * EXT2_INODE_SIZE(&sb));
 #ifdef DEBUG
-	printf("ext2_iget: reading %ld bytes at offset %ld "
-	       "(%ld + (%d * %d) + ((%d) %% %d) * %d) "
+	printf("ext2_iget: reading %ld bytes at offset %lld "
+	       "(%lld + (%d * %d) + ((%d) %% %d) * %d) "
 	       "(inode %d -> table %d)\n", 
 	       sizeof(struct ext2_inode), offset, partition_offset,
 	       ext2_gds(group)->bg_inode_table, ext2_blocksize,
@@ -493,7 +519,7 @@ static int ext3_extent_load_find(struct ext2_inode *ip, int leaf, int d,
 			printf("load extent tree[%d] block at %d\n", d, leaf);
 
 		if (cons_read(dev, hdr, EXT2_MAX_BLOCK_SIZE,
-			      leaf * ext2_blocksize) !=
+			      mul_blocksize(leaf, ext2_blocksize)) !=
 		    EXT2_MAX_BLOCK_SIZE) {
 			printf("ext3_extent_load_find: read error\n");
 			return -1;
@@ -555,7 +581,7 @@ static int ext2_blkno(struct ext2_inode *ip, int blkoff)
 	int blkno;
 	int iblkno;
 	int diblkno;
-	unsigned long offset;
+	__u64 offset;
 
 	if (ip->i_flags & EXT3_EXTENTS_FL)
 		return ext3_extent_blkno(ip, blkoff);
@@ -578,7 +604,7 @@ static int ext2_blkno(struct ext2_inode *ip, int blkoff)
 
 		/* Read the indirect block */
 		if (cached_iblkno != iblkno) {
-			offset = partition_offset + (long)iblkno * (long)ext2_blocksize;
+			offset = partition_offset + mul_blocksize(iblkno, ext2_blocksize);
 			if (cons_read(dev, iblkbuf, ext2_blocksize, offset)
 			    != ext2_blocksize)
 			{
@@ -605,7 +631,7 @@ static int ext2_blkno(struct ext2_inode *ip, int blkoff)
 
 		/* Read in the double-indirect block */
 		if (cached_diblkno != diblkno) {
-			offset = partition_offset + (long) diblkno * (long) ext2_blocksize;
+			offset = partition_offset + mul_blocksize(diblkno, ext2_blocksize);
 			if (cons_read(dev, diblkbuf, ext2_blocksize, offset)
 			    != ext2_blocksize)
 			{
@@ -626,7 +652,7 @@ static int ext2_blkno(struct ext2_inode *ip, int blkoff)
 		/* Read the indirect block */
     
 		if (cached_iblkno != iblkno) {
-			offset = partition_offset + (long) iblkno * (long) ext2_blocksize;
+			offset = partition_offset + mul_blocksize(iblkno, ext2_blocksize);
 			if (cons_read(dev, iblkbuf, ext2_blocksize, offset)
 			    != ext2_blocksize)
 			{
@@ -662,7 +688,8 @@ static int print_ext2_blkno(struct ext2_inode *ip, int blkoff)
 static int ext2_breadi(struct ext2_inode *ip, long blkno, long nblks,
 		       char *buffer)
 {
-	long dev_blkno, ncontig, offset, nbytes, tot_bytes;
+	long dev_blkno, ncontig, nbytes, tot_bytes;
+	__u64 offset;
 
 	if (Debug) printf("ext2_breadi(%p, %ld, %ld, %p)\n",
 		ip, blkno, nblks, buffer);
@@ -693,9 +720,9 @@ static int ext2_breadi(struct ext2_inode *ip, long blkno, long nblks,
 			memset(buffer, 0, nbytes);
 		} else {
 			/* Read it for real */
-			offset = partition_offset + (long) dev_blkno* (long) ext2_blocksize;
+			offset = partition_offset + mul_blocksize(dev_blkno, ext2_blocksize);
 #ifdef DEBUG
-			printf("ext2_bread: reading %ld bytes at offset %ld\n",
+			printf("ext2_bread: reading %ld bytes at offset %lld\n",
 			       nbytes, offset);
 #endif
 			if (cons_read(dev, buffer, nbytes, offset)
@@ -750,7 +777,7 @@ static struct ext2_dir_entry_2 *ext2_readdiri(struct ext2_inode *dir_inode,
 	}
 
 	dp = (struct ext2_dir_entry_2 *) (blkbuf + blockoffset);
-	swapde(dp);
+	dp = swapde(dp);
 	blockoffset += dp->rec_len;
 	/* ext2 deletes a file by zeroing its inode.  We skip deleted
 	 * files, corrupt entries and entries that aren't a regular
@@ -857,10 +884,12 @@ static struct ext2_inode *ext2_namei(const char *name)
  * Note: don't mix any kind of file lookup or other I/O with this or
  * you will lose horribly (as it reuses blkbuf)
  */
-const char * ext2_readdir(int fd, int rewind)
+const char * ext2_readdir(int fd, int rewind, unsigned char *file_type)
 {
 	struct ext2_inode * ip = fd2inode[fd];
 	struct ext2_dir_entry_2 * ent;
+	if (file_type)
+		*file_type = 0;
 	if (!S_ISDIR(ip->i_mode)) {
 		printf("fd %d (inode %d) is not a directory (mode %x)\n",
 		       fd, inode_table[fd].inumber, ip->i_mode);
@@ -869,6 +898,8 @@ const char * ext2_readdir(int fd, int rewind)
 	ent = ext2_readdiri(ip, rewind);
 	if (ent) {
 		ent->name[ent->name_len] = '\0';
+		if (file_type)
+			*file_type = ent->file_type;
 		return ent->name;
 	} else { 
 		return NULL;
@@ -951,16 +982,16 @@ static struct ext2_inode * ext2_follow_link(struct ext2_inode * from,
 	}
 }
 
-static int ext2_read(int fd, char *buf, unsigned count, unsigned devaddr)
+static int ext2_read(int fd, char *buf, unsigned count, __u64 devaddr)
 {
 	struct ext2_inode * ip;
 	ip = fd2inode[fd];
 
-	if (Debug) printf("ext2_read(%d, 0x%p, %d, %d)\n",
+	if (Debug) printf("ext2_read(%d, 0x%p, %d, %lld)\n",
 		fd, buf, count, devaddr);
 
 	return ext2_breadi(ip,
-		devaddr / ext2_blocksize,
+		div_blocksize(devaddr, ext2_blocksize),
 		count / ext2_blocksize,
 		buf);
 }
@@ -1004,4 +1035,6 @@ void ext2_close(int fd)
 	/* blah, hack, don't close the root inode ever */
 	if (&inode_table[fd].inode != root_inode)
 		ext2_iput(&inode_table[fd].inode);
+
+	fileio_close(fd);
 }
