@@ -19,6 +19,7 @@ extern char _end, _start, _edata;
 
 int Debug = 0;
 int interactive = 0;
+int disk_2gb_limit;
 
 /*
  * Beware: pdc_model_sysmodel() may return a machine name which has trailing
@@ -200,7 +201,7 @@ load_kernel(int fd, unsigned *entryp, int *wide)
     for (i = 0; i < loadable.n; i++)
     {
 	loadable.segment[i].mem =  PHYS(loadable.segment[i].mem);
-	printf("Segment %d load %08x size %d mediaptr 0x%lx\n",
+	printf("Segment %d load 0x%08x size %d mediaptr 0x%lx\n",
 		i, loadable.segment[i].mem, loadable.segment[i].length,
 		loadable.segment[i].offset);
 
@@ -310,6 +311,7 @@ ls(char *path)
     char  *p, kern_dir[256];
     const char *dir;
     int fd, part, part_fd;
+    unsigned char file_type;
 
     parse_pfname(path, &part, kern_dir);
     if ((p = strrchr(kern_dir, '/')) != NULL)
@@ -322,29 +324,34 @@ ls(char *path)
 	strcpy(kern_dir, "/.");
     }
 
-    printf("Directory listing of %s\n\n", kern_dir);
+    printf("Directory listing of %d%s\n\n", part, kern_dir);
 
     partition_transform(&part, NULL);
 
     /* partition table starts from zero */
-    part_fd = offset_open(bootdev, 512 * partition[part - 1].start,
-			      512 * partition[part - 1].length);
+    part_fd = offset_open(bootdev, 512ULL * partition[part - 1].start,
+			      512ULL * partition[part - 1].length);
     if(ext2_mount(part_fd, 0, 0) == -1) {
 	printf("Failed to mount partition %d\n", part);
+	fileio_close(part_fd);
 	return;
     }
 
     if((fd = ext2_open(kern_dir)) == -1) {
 	printf("Failed to open directory %s\n", kern_dir);
+	fileio_close(part_fd);
 	return;
     }
 
-    while((dir = ext2_readdir(fd, 0)) != NULL)
+    while((dir = ext2_readdir(fd, 0, &file_type)) != NULL)
 	if(dir[0] != '.') /* skip hidden files and . and .. */
-	    printf("  %s\n", dir);
+	    printf("  %s%s\n", dir,
+		file_type == EXT2_FT_SYMLINK ? "@" :
+		file_type == EXT2_FT_DIR ? "/" : "");
     
     printf("\n");
     ext2_close(fd);
+    fileio_close(part_fd);
     /* There's no umount ... since the next mount reuses
      * the current one's data structures ... */
 }
@@ -356,7 +363,7 @@ interact(int *ok)
     char *argv[MAX_ARGV], *p;
     char orig[CMDLINELEN];
     const char sep[] = " \t";
-    char numbuf[4];
+    char numbuf[32];
     char fieldbuf[200];
     int i, argc, editfield;
 
@@ -390,13 +397,13 @@ interact(int *ok)
 	    printf("%2d: %s\n", i, argv[i]);
 	}
 	printf("\n"
-	    "<#>    edit the numbered field\n"
-	    "'b'    boot with this command line\n"
-	    "'r'    restore command line\n"
-	    "'l'    list dir\n"
-	    "'x'    reset and reboot machine\n"
+	    "<#>           edit the numbered field\n"
+	    "'b'           boot with this command line\n"
+	    "'r'           restore command line\n"
+	    "'l <#/path>   list dir from numbered partition with path\n"
+	    "'x'           reset and reboot machine\n"
 #ifdef HIDDEN_COMMAND
-	    "'d'    toggle debug mode\n"
+	    "'d'           toggle debug mode\n"
 #endif
 	    "? ");
 	numbuf[0] = '0';
@@ -419,7 +426,14 @@ interact(int *ok)
 	if (numbuf[0] == 'l')
 	{
 	    join(commandline, argc, argv, ok);
-	    ls(argv[0]);
+	    if (numbuf[1] == ' ') {
+		/* add trailing slash */
+		p = strrchr(numbuf, '/');
+		if (p && p[1] != 0)
+			strcat(numbuf, "/");
+		ls(&numbuf[2]);
+	    } else
+		ls(argv[0]);
 	    continue;
 	}
 
@@ -529,6 +543,9 @@ iplmain(int is_interactive, char *initialstackptr, int started_wide)
 	printf("palo loaded at %p-%p.\n", &_start, &_end);
     }
 
+    disk_2gb_limit = pdc_bootdisk_2GB_limit();
+    printf("Boot limited to sectors below 2GB: %s\n", disk_2gb_limit ? "YES":"NO");
+
  restart:
 
     blocked_bootdev = pdc_bootdev_open();
@@ -730,7 +747,8 @@ iplmain(int is_interactive, char *initialstackptr, int started_wide)
 
 	    kernfd = -1;
 
-	    if ((pdc_os_bits() & (OS_32|OS_64)) == (OS_32|OS_64))
+	    if ((pdc_os_bits() & (PDC_MODEL_OS32|PDC_MODEL_OS64))
+			== (PDC_MODEL_OS32|PDC_MODEL_OS64))
 	    {
 		printf("\nThis box can boot either 32 or 64-bit kernels...\n");
 		if (f.kern32_offset == 0 && f.kern64_offset != 0)
@@ -755,12 +773,12 @@ iplmain(int is_interactive, char *initialstackptr, int started_wide)
 		}
 	    }
 
-	    if (kernfd == -1 && (pdc_os_bits() & OS_64))
+	    if (kernfd == -1 && (pdc_os_bits() & PDC_MODEL_OS64))
 	    {
 		printf("Choosing 64-bit kernel\n");
 		kernfd = offset_open(bootdev, f.kern64_offset, f.kern64_sz);
 	    }
-	    else if (kernfd == -1 && (pdc_os_bits() & OS_32))
+	    else if (kernfd == -1 && (pdc_os_bits() & PDC_MODEL_OS32))
 	    {
 		printf("Choosing 32-bit kernel\n");
 		kernfd = offset_open(bootdev, f.kern32_offset, f.kern32_sz);
@@ -769,6 +787,7 @@ iplmain(int is_interactive, char *initialstackptr, int started_wide)
 
 	if (!load_kernel(kernfd, &entry, &wide))
 	{
+	    fileio_close(kernfd);
 	    die("ERROR: failed to load kernel\n");
 	    goto restart;
 	}
@@ -780,6 +799,7 @@ iplmain(int is_interactive, char *initialstackptr, int started_wide)
 	    {
 		printf("ERROR: failed to load ramdisk - proceeding anyway\n");
 	    }
+	    fileio_close(rdfd);
 	}
     }
     else /* kern_part > 0 && we're partitioned */
@@ -802,7 +822,7 @@ iplmain(int is_interactive, char *initialstackptr, int started_wide)
 
 	pp = &partition[kern_part - 1];
     
-	part_fd = offset_open(bootdev, 512 * pp->start, 512 * pp->length);
+	part_fd = offset_open(bootdev, 512ULL * pp->start, 512ULL * pp->length);
 
 	mount_fd = ext2_mount(part_fd, 0, 0);
 	if (0) printf("ext2_mount(partition %d) returns %d\n",
@@ -814,12 +834,17 @@ iplmain(int is_interactive, char *initialstackptr, int started_wide)
 	{
 	    printf("ERROR: open %s from partition %d failed\n",
 		kern_name, kern_part);
+	    fileio_close(mount_fd);
+	    fileio_close(part_fd);
 	    goto restart;
 	}
 
 	bkern_fd = byteio_open(kern_fd);
 	if (!load_kernel(bkern_fd, &entry, &wide))
 	{
+	    fileio_close(bkern_fd);
+	    fileio_close(mount_fd);
+	    fileio_close(part_fd);
 	    die("ERROR: failed to load kernel\n");
 	    goto restart;
 	}
