@@ -9,7 +9,7 @@
  */
 
 #define PALOMAGIC "PALO"
-#define PALOVERSION "2.15"
+#define PALOVERSION "2.22"
 
 #ifndef __ASSEMBLY__
 
@@ -134,16 +134,18 @@ struct firstblock
 struct diskpartition
 {
     /* in honor of the MBR scheme, these are in 512-byte sectors */
-    unsigned start;
+    __u64 start;
     /* length == 0 means nothing's here */
-    unsigned length;
+    __u64 length;
     unsigned char id;
 };
 
 
 #define MAXPARTS 16
 
-int seekread(int fd, char *buf, unsigned nbytes, unsigned devaddr);
+extern int disk_2gb_limit;
+
+int seekread(int fd, char *buf, unsigned nbytes, __u64 devaddr);
 #define STRUCTWRITE(f, data, where) seekwrite((f), (char *)&data, sizeof data, (where))
 #define STRUCTREAD(f, data, where) seekread((f), (char *)&data, sizeof data, (where))
 #define GZIP_STRUCTREAD(loadable, f, data, where)			\
@@ -155,7 +157,6 @@ int seekread(int fd, char *buf, unsigned nbytes, unsigned devaddr);
 /* diskpart.c */
 extern int is_extended(int id);
 extern int load_partitions(int bootdev, struct diskpartition *mptab, int maxparts);
-extern void print_ptab(struct diskpartition *mptab, int maxparts);
 extern void print_ptab_pretty(struct diskpartition *mptab, int maxparts);
 
 #endif /* __ASSEMBLY__ */
